import asyncio
from asyncio import AbstractEventLoop
from typing import Optional, List, Callable, Union

import websockets
from loguru import logger

try:
    from orjson import orjson as json
except ImportError:
    import json

from dalas.types import CentrifugoChannel
from dalas.http_requests import BaseRequester
from dalas.models.data_models import UserInfo, NewAlertRaw
from dalas.models.raw_models import ChannelSubRaw


class FromFuncHandler:
    def __init__(self, handler: Callable, channel: CentrifugoChannel, blocking: bool = True):
        self.handler = handler
        self.channel = channel
        self.blocking = blocking

    async def check(self, event: NewAlertRaw):
        return event.result.channel.startswith(self.channel.value)


class Alerts:
    def __init__(self,
                 user: UserInfo,
                 api_requester: Optional[BaseRequester],
                 loop: Optional[AbstractEventLoop] = None):
        self.user = user
        self.__request_api = api_requester.request_api
        self.__WS_URL = 'wss://centrifugo.donationalerts.com/connection/websocket'

        self.loop = loop or asyncio.new_event_loop()

        self.handlers: List[FromFuncHandler] = []
        self.channels: List[CentrifugoChannel] = []

    async def subscribe(self):

        connect_json = json.dumps({'params': {'token': self.user.socket_connection_token},
                                   'id': self.user.id})

        async with websockets.connect(self.__WS_URL) as ws:
            await ws.send(connect_json)

            channel_response = ChannelSubRaw(**json.loads(await ws.recv()))

            api_response = await self.__request_api('centrifuge/subscribe',
                                                    method='POST',
                                                    channels=[f'{channel.value}_{channel_response.id}' for channel in
                                                              self.channels],
                                                    client=channel_response.result.client)
            for i in range(len(api_response["channels"])):
                await ws.send(json.dumps({
                    "params": {
                        "channel": api_response["channels"][i]["channel"],
                        "token": api_response["channels"][i]["token"]
                    },
                    "method": 1,
                    "id": self.user.id
                }))

            await self.listen_websocket(ws)

    async def listen_websocket(self, ws):
        logger.info('Websocket listener started')
        async for event in ws:
            js = json.loads(event)
            if 'id' in js.keys() or 'type' in js['result'].keys():
                continue

            event = NewAlertRaw(**js)
            logger.debug(f"Websocket received new event: {event.json()}")
            for handler in self.handlers:
                result = await handler.check(event)
                logger.debug(f'Handler {handler.handler.__name__} returned <{result}>')
                if result:
                    await handler.handler(event.result.data.data)
                if handler.blocking:
                    break
        logger.info('Websocket listener finished')

    def new_alert(self, channel: CentrifugoChannel, blocking=True):
        def decorator(func):
            self.handlers.append(
                FromFuncHandler(func, blocking=blocking, channel=channel)
            )
            self.channels.append(channel)
            return func

        return decorator



from abc import ABC

import aiohttp
from loguru import logger
from .validators import BaseValidator

try:
    from orjson import orjson as json
except ImportError:
    import json


class BaseRequester(ABC):
    def __init__(self, base_url: str, token: str, validator: BaseValidator):
        self.__BASE_URL = base_url
        self.__token = token
        self.validator = validator

    async def request_api(self, method_name: str, method: str = 'GET', **kwargs):
        raise NotImplementedError()


class HTTPRequester(BaseRequester):
    def __init__(self, base_url: str, token: str, validator: BaseValidator):
        super().__init__(base_url, token, validator)
        self.__BASE_URL = base_url
        self.__token = token
        self.validator = validator

    async def request_api(self, method_name: str, method: str = 'GET', **kwargs):
        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False)) as session:
            async with session.request(url=self.__BASE_URL + method_name,
                                       method=method, data=json.dumps(kwargs),
                                       headers={"Authorization": f"Bearer {self.__token}",
                                                "Content-Type": "application/json"}) as request:
                response = await request.json()
                logger.debug(f'Requesting "{method_name}" with {kwargs} returned: {response}')

        response = await self.validator.response_validator(response)
        logger.debug("API response was validated")
        return response



from typing import Union, List, Optional, Dict

from dalas.http_requests import HTTPRequester, BaseRequester
from dalas.models.data_models import UserInfo, Donation, SendCustomAlerts
from dalas.models.raw_models import UserInfoRaw, DonationAlertsListRaw, SendCustomAlertsRaw
from dalas.types import Locales
from dalas.validators import BaseValidator, DefaultValidator

MerchTitle = Dict[Union[str, Locales], str]


class API:

    def __init__(self,
                 api_requester: Optional[BaseRequester],
                 validator: BaseValidator = DefaultValidator(),
                 raw_responses: bool = False):
        self.__request_api = api_requester.request_api
        self.validator = validator
        self.raw_responses = raw_responses

    async def user(self) -> Union[UserInfoRaw, UserInfo]:
        response: UserInfoRaw = UserInfoRaw(**(await self.__request_api('user/oauth')))
        if self.raw_responses:
            return response
        return response.data

    async def donation_alerts_list(self) -> Union[DonationAlertsListRaw, List[Donation]]:
        response: DonationAlertsListRaw = DonationAlertsListRaw(**(await self.__request_api('alerts/donations')))
        if self.raw_responses:
            return response
        return response.data

    async def send_custom_alerts(self,
                                 external_id: Optional[str] = None,
                                 header: Optional[str] = None,
                                 message: Optional[str] = None,
                                 is_shown: Optional[Union[int, bool]] = None,
                                 image_url: Optional[str] = None,
                                 sound_url: Optional[str] = None,
                                 **kwargs) -> Union[SendCustomAlerts, SendCustomAlertsRaw]:
        params = await self.validator.request_validator(locals())
        response: SendCustomAlertsRaw = SendCustomAlertsRaw(
            **(await self.__request_api('custom_alert', method='POST', **params)))
        if self.raw_responses:
            return response
        return response.data

    # todo: finish when possible
    # async def create_merch(self,
    #                        merchant_identifier: str,
    #                        merchandise_identifier: str,
    #                        title: MerchTitle,
    #                        currency: Union[str, InputCurrencies],
    #                        price_user: Union[float, int],
    #                        price_service: Union[float, int],
    #                        signature: str,
    #                        end_at_ts: int = int(time.time()),
    #                        is_active: Union[int, bool] = False,
    #                        is_percentage: Union[int, bool] = False,
    #                        url: Optional[str] = None,
    #                        img_url: Optional[str] = None):
    #     params = await self.validator.request_validator(locals())
    #     return await self.__request_api('merchandise', method='POST', **params)
    #
    # async def update_merch(self,
    #                        id: str,
    #                        merchant_identifier: Optional[str] = None,
    #                        merchandise_identifier: Optional[str] = None,
    #                        title: Optional[Union[MerchTitle, dict]] = None,
    #                        is_active: Optional[Union[bool, int]] = None,
    #                        is_percentage: Optional[Union[bool, int]] = None,
    #                        currency: Optional[Union[InputCurrencies, str]] = None,
    #                        price_user: Union[float, int] = None,
    #                        price_service: Union[float, int] = None,
    #                        url: Optional[str] = None,
    #                        img_url: Optional[str] = None,
    #                        end_at_ts: Optional[int] = int(time.time()),
    #                        signature: Optional[str] = None):
    #
    #     params = await self.validator.request_validator(locals())
    #     return await self.__request_api(f'merchandise/{id}', method='PUT', **params)
    #
    # async def send_sale_alerts(self,
    #                            user_id: int,
    #                            external_id: str,
    #                            merchant_identifier: str,
    #                            merchandise_identifier: str,
    #                            amount: Union[float, int],
    #                            currency: Union[str, InputCurrencies],
    #                            signature: str,
    #                            bought_amount: int = 1,
    #                            username: Optional[str] = None,
    #                            message: Optional[str] = None):
    #
    #     params = await self.validator.request_validator(locals())
    #     return await self.__request_api('merchandise_sale', method='POST', **params)



import asyncio
from asyncio import AbstractEventLoop
from typing import Optional, Union

from .alerts import Alerts
from .api import API
from .http_requests import BaseRequester, HTTPRequester

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from loguru import logger

from .token import Token
from .validators import BaseValidator, DefaultValidator


class Dalas:

    def __init__(self,
                 token: Union[str, Token],
                 client_id: int = None,
                 scopes: Optional[Union[str, list]] = None,
                 path: Optional[str] = None,
                 validator: BaseValidator = DefaultValidator(),
                 api_requester: Optional[BaseRequester] = None,
                 raw_responses: bool = False,
                 loop: Optional[AbstractEventLoop] = None):

        if token and any((client_id, scopes, path)):
            logger.warning('Token passed, other args are ignored.')

        if isinstance(token, str):
            self.__token = token
        elif isinstance(token, Token):
            self.__token = asyncio.run(token.key)
        else:
            raise ValueError("No token or token generator passed")

        self.loop = loop or asyncio.new_event_loop()
        self.__BASE_URL = 'https://www.donationalerts.com/api/v1/'

        self.api_requester = api_requester or HTTPRequester(self.__BASE_URL, self.__token, validator)

        self.api = API(self.api_requester, validator, raw_responses)
        self.alerts = Alerts(self.user, self.api_requester, self.loop)

    @property
    def user(self):
        return self.loop.run_until_complete(self.api.user())



from abc import ABC, abstractmethod
import enum


class BaseValidator(ABC):

    @abstractmethod
    async def request_validator(self, request):
        pass

    @abstractmethod
    async def response_validator(self, response):
        pass


class DefaultValidator(BaseValidator):

    async def response_validator(self, response: dict):
        response_copy = response.copy()
        for k, v in response.items():
            if k.startswith('is_') and isinstance(v, int):
                response_copy[k] = bool(v)
            elif k in ('from',
                       'to'):  # we need this horrible piece of shit bc name "from" reserved, but we have "from" in "links". "to" renamed just for compatibility
                del response_copy[k]
                response_copy[f'{k}_index'] = v
            elif isinstance(v, dict):
                response_copy[k] = await self.response_validator(v)
            elif isinstance(v, list):
                values = []
                for i in range(len(v)):
                    values.append(await self.response_validator(v[i]))
                response_copy[k] = values

        return response_copy

    async def request_validator(self, request: dict):
        request_copy = request.copy()
        for k, v in request.items():

            if isinstance(k, enum.Enum):
                del request_copy[k]
                k = k.value
                request_copy[k] = v
            elif k == 'self' or k.startswith('__') or not v:
                del request_copy[k]

            elif isinstance(v, bool):
                request_copy[k] = int(v)
            elif isinstance(v, enum.Enum):
                request_copy[k] = v.value
            elif isinstance(v, dict):
                request_copy[k] = await self.request_validator(v)
        return request_copy



from setuptools import setup

setup(
    name='DaLaS',
    version='0.1.0',
    packages=['dalas'],
    url='https://github.com/TheRealMizAL/DaLaS',
    license='GNU GPL v3.0',
    author='kov20',
    author_email='amizurenko2002@gmail.com',
    description='Fully customizable asynchronous Donation Alerts API',
    install_requires=[
        'pydantic>=1.9.1',
        'loguru>=0.6.0',
        'PyYAML>=6.0',
        'websockets>=10.3',
        'aiohttp>=3.8.1'
    ]
)



from typing import Union, Optional
from urllib.parse import urlencode
import time

import webbrowser
import yaml

from .http_requests import HTTPRequester
from .types import TokenRequest, TokenResponse
from .validators import DefaultValidator


class BaseToken:
    def __init__(self, path: str,
                 redirect_uri: str,
                 client_id: int,
                 response_type: str,
                 scope: Union[str, list]):
        self.path = path or '~/.donationalerts/secret.yaml'
        self.redirect_uri = redirect_uri
        self.client_id = client_id
        self.response_type = response_type
        self.scope = scope

    async def auth_request(self):
        data: TokenRequest = TokenRequest(client_id=self.client_id, redirect_uri=self.redirect_uri,
                                          response_type=self.response_type, scope=self.scope)
        url = 'https://www.donationalerts.com/oauth/authorize?' + urlencode(data.dict())
        webbrowser.open(url)

        return input(f'Copy {self.response_type} from url and paste here: ')

    @property
    async def key(self):
        raise NotImplementedError()


class AuthCode(BaseToken):
    def __init__(self,
                 path: str,
                 redirect_uri: str,
                 client_id: int,
                 scope: Union[str, list],
                 client_secret: str):
        super().__init__(path, redirect_uri, client_id, 'code', scope)
        self.client_secret = client_secret

    @property
    async def key(self):
        try:
            with open(self.path, 'r', encoding='utf-8') as file:
                token_response = TokenResponse(**yaml.load(file, yaml.Loader))
        except FileNotFoundError:
            token_response = await self.__get_token()
            with open(self.path, 'w', encoding='utf-8') as file:
                yaml.dump(token_response.dict(), file)
            return token_response.access_token

        if token_response.expires_in < int(time.time()):
            token_response = await self.__refresh_token(token_response.refresh_token)
            with open(self.path, 'w', encoding='utf-8') as file:
                yaml.dump(token_response.dict(), file)

        return token_response.access_token

    async def __get_token(self):
        code = await self.auth_request()
        response = await HTTPRequester('https://www.donationalerts.com/', '', DefaultValidator(), ).request_api(
            'oauth/token',
            'POST',
            grant_type='authorization_code',
            client_id=self.client_id,
            client_secret=self.client_secret,
            redirect_uri=self.redirect_uri,
            code=code)

        return TokenResponse(**response)

    async def __refresh_token(self, refresh_token: str):
        response = await HTTPRequester('https://www.donationalerts.com/', '', DefaultValidator()).request_api(
            'oauth/token',
            'POST',
            grant_type='refresh_token',
            refresh_token=refresh_token,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope)

        return TokenResponse(**response)


class Implicit(BaseToken):

    def __init__(self,
                 path: str,
                 redirect_uri: str,
                 client_id: int,
                 scope: Union[str, list]):
        super().__init__(path, redirect_uri, client_id, 'token', scope)

    @property
    async def key(self):
        try:
            with open(self.path, 'r', encoding='utf-8') as file:
                return yaml.load(file, yaml.Loader)['access_token']
        except FileNotFoundError:
            token_response = await self.auth_request()
            with open(self.path, 'w', encoding='utf-8') as file:
                yaml.dump({'access_token': token_response}, file)
            return token_response


class Token:
    def __init__(self,
                 client_id: int,
                 scope: Union[str, list],
                 path: str = '~/.donationalerts/secret.yaml',
                 redirect_uri: str = '127.0.0.1:5000/',
                 response_type: str = 'token',
                 client_secret: Optional[str] = None):
        if response_type == 'token':
            self.token = Implicit(path, redirect_uri, client_id, scope)
        elif response_type == 'code':
            if not client_secret:
                raise ValueError('client_secret can\'t be None')
            self.token = AuthCode(path, redirect_uri, client_id, scope, client_secret)
        else:
            raise ValueError('Wrong response_type')

    @property
    async def key(self):
        return await self.token.key



import enum
from pydantic import BaseModel

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Scopes:
    OAUTH_USER_SHOW = 'oauth-user-show'
    OUATH_DONATION_SUBSCRIBE = 'oauth-donation-subscribe'
    OAUTH_DONATION_INDEX = 'oauth-donation-index'
    OAUTH_CUSTOM_ALERT_STORE = 'oauth-custom_alert-store'
    OAUTH_GOAL_SUBSCRIBE = 'oauth-goal-subscribe'
    OAUTH_POLL_SUBSCRIBE = 'oauth-poll-subscribe'

    ALL = ' '.join([OAUTH_POLL_SUBSCRIBE, OAUTH_GOAL_SUBSCRIBE, OAUTH_CUSTOM_ALERT_STORE, OUATH_DONATION_SUBSCRIBE,
                    OAUTH_DONATION_INDEX, OAUTH_USER_SHOW])


class Locales(enum.Enum):
    Belarusian = 'be_BY'
    German = 'de_DE'
    English_US = 'en_US'
    Spanish = 'es_ES'
    Spanish_US = 'es_US'
    Estonian = 'et_EE'
    French = 'fr_FR'
    Hebrew = 'he_HE'
    Italian = 'it_IT'
    Georgian = 'ka_GE'
    Kazakh = 'kk_KZ'
    Korean = 'ko_KR'
    Latvian = 'lv_LV'
    Polish = 'pl_PL'
    Portuguese_BR = 'pt_BR'
    Russian = 'ru_RU'
    Swedish = 'sv_SE'
    Turkish = 'tr_TR'
    Ukrainian = 'uk_UA'
    Chinese = 'zh_CN'


class InputCurrencies(enum.Enum):
    EURO = 'EUR'
    US_DOLLAR = 'USD'
    RU_RUBLE = 'RUB'
    REAL = 'BRL'
    LIRA = 'TRY'


class OutputCurrencies(enum.Enum):
    EURO = 'EUR'
    US_DOLLAR = 'USD'
    RU_RUBLE = 'RUB'
    REAL = 'BRL'
    LIRA = 'TRY'
    BU_RUBLE = 'BYN'
    TENGE = 'KZT'
    HRYVNIA = 'UAH'


class CentrifugoChannel(enum.Enum):
    NEW_DONATION_ALERTS = "$alerts:donation"
    DONATION_GOALS_UPDATES = "$goals:goal"
    POLLS_UPDATES = "$polls:poll"


class TokenRequest(BaseModel):
    client_id: int
    redirect_uri: str
    response_type: Literal["code", "token"]
    scope: str


class TokenResponse(BaseModel):
    token_type: str
    expires_in: int
    access_token: str
    refresh_token: str
